# Adopted from tatsu-lab@stanford_alpaca. Below is the original copyright:
#    Copyright 2023 Rohan Taori, Ishaan Gulrajani, Tianyi Zhang, Yann Dubois, Xuechen Li
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from loguru import logger
import copy
from dataclasses import dataclass, field
import json
import pathlib
from typing import Dict, Optional, Sequence
from datasets import load_dataset
from itertools import chain
from glob import glob
import os


import numpy as np
import torch
from torch.utils.data import Dataset
import transformers
from transformers import Trainer
from transformers.trainer_pt_utils import LabelSmoother

from fastchat.conversation import SeparatorStyle
from fastchat.model.model_adapter import get_conversation_template

IGNORE_TOKEN_ID = LabelSmoother.ignore_index


@dataclass
class ModelArguments:
    model_name_or_path: Optional[str] = field(default="facebook/opt-125m")


@dataclass
class DataArguments:
    dataset_name: Optional[str] = field(
        default=None, metadata={"help": "The name of the dataset to use (via the datasets library)."}
    )
    dataset_config_name: Optional[str] = field(
        default=None, metadata={"help": "The configuration name of the dataset to use (via the datasets library)."}
    )
    train_file_dir: Optional[str] = field(default=None, metadata={"help": "The train text data file folder."})
    validation_file_dir: Optional[str] = field(
        default=None,
        metadata={"help": "An optional input evaluation data file to evaluate the perplexity on text file folder."},
    )
    max_train_samples: Optional[int] = field(
        default=None,
        metadata={
            "help": (
                "For debugging purposes or quicker training, truncate the number of training examples to this "
                "value if set."
            )
        },
    )
    max_eval_samples: Optional[int] = field(
        default=None,
        metadata={
            "help": (
                "For debugging purposes or quicker training, truncate the number of evaluation examples to this "
                "value if set."
            )
        },
    )
    streaming: bool = field(default=False, metadata={"help": "Enable streaming mode"})
    block_size: Optional[int] = field(
        default=2048,
        metadata={
            "help": (
                "Optional input sequence length after tokenization. "
                "The training dataset will be truncated in block of this size for training. "
                "Default to the model max input length for single sentence inputs (take into account special tokens)."
            )
        },
    )
    overwrite_cache: bool = field(
        default=False, metadata={"help": "Overwrite the cached training and evaluation sets"}
    )
    validation_split_percentage: Optional[float] = field(
        default=1,
        metadata={
            "help": "The percentage of the train set used as validation set in case there's no validation split"
        },
    )
    preprocessing_num_workers: Optional[int] = field(
        default=128,
        metadata={"help": "The number of processes to use for the preprocessing."},
    )
    keep_linebreaks: bool = field(
        default=True, metadata={"help": "Whether to keep line breaks when using TXT files or not."}
    )


@dataclass
class TrainingArguments(transformers.TrainingArguments):
    cache_dir: Optional[str] = field(default=None)
    optim: str = field(default="adamw_torch")
    model_max_length: int = field(
        default=2048,
        metadata={
            "help": "Maximum sequence length. Sequences will be right padded (and possibly truncated)."
        },
    )
    disable_tqdm: bool = False
    adam_beta1: float = 0.9
    adam_beta2: float = 0.95
    weight_decay: float = 0.1

local_rank = None


def rank0_print(*args):
    if local_rank == 0:
        print(*args)


def safe_save_model_for_hf_trainer(trainer: transformers.Trainer, output_dir: str):
    """Collects the state dict and dump to disk."""
    state_dict = trainer.model.state_dict()
    if trainer.args.should_save:
        cpu_state_dict = {key: value.cpu() for key, value in state_dict.items()}
        del state_dict
        trainer._save(output_dir, state_dict=cpu_state_dict)  # noqa




def get_dataset(tokenizer, data_args, training_args):
    
    def tokenize_function(examples):
        # input_ids = tokenizer(
        #     examples["abstract"],
        #     return_tensors="pt",
        #     padding="max_length",
        #     max_length=tokenizer.model_max_length,
        #     truncation=True,
        # ).input_ids
        # targets = input_ids.clone()

        # return  dict(
        #     input_ids=input_ids,
        #     labels=targets, 
        #     attention_mask=input_ids.ne(tokenizer.pad_token_id),
        # )
        return tokenizer(examples["text"])

    # Main data processing function that will concatenate all texts from our dataset and generate chunks of block_size.
    def group_texts(examples):
        # Concatenate all texts.
        concatenated_examples = {k: list(chain(*examples[k])) for k in examples.keys()}
        total_length = len(concatenated_examples[list(examples.keys())[0]])
        # We drop the small remainder, we could add padding if the model supported it instead of this drop, you can
        # customize this part to your needs.
        if total_length >= tokenizer.model_max_length:
            total_length = (total_length // tokenizer.model_max_length) * tokenizer.model_max_length
        # Split by chunks of max_len.
        result = {
            k: [t[i: i + tokenizer.model_max_length] for i in range(0, total_length, tokenizer.model_max_length)]
            for k, t in concatenated_examples.items()
        }

        # for i in range(len(result['input_ids'])):
        #     result['input_ids'][i] = [259 if x == 1 else x for x in result['input_ids'][i]]
        #     result['input_ids'][i][0] = 1


            # print(tokenizer.decode(result['input_ids'][i]))
        # try:
        #     result['input_ids'].to('cuda')
        # except:
        #     print('err s1gh')
        # print(tokenizer.decode(result['input_ids']))
        # print(len(result['input_ids']))
        result["labels"] = result["input_ids"].copy()
        # print(result)
        # print(len(examples["input_ids"]))
        # print(len(examples["labels"]))
        # print(len(examples['attention_mask']))
        # print(result.keys())
        return result


    # Get the datasets: you can either provide your own CSV/JSON/TXT training and evaluation files (see below)
    # or just provide the name of one of the public datasets available on the hub at https://huggingface.co/datasets/
    # (the dataset will be downloaded automatically from the datasets Hub).
    #
    # For CSV/JSON files, this script will use the column called 'text' or the first column if no column called
    # 'text' is found. You can easily tweak this behavior (see below).
    #
    # In distributed training, the load_dataset function guarantee that only one local process can concurrently
    # download the dataset.

    data_files = {}
    dataset_args = {}
    if data_args.train_file_dir is not None and os.path.exists(data_args.train_file_dir):
        train_data_files = glob(f'{data_args.train_file_dir}/**/*.txt', recursive=True)
        logger.info(f"train files: {', '.join(train_data_files)}")
        data_files["train"] = train_data_files
    if data_args.validation_file_dir is not None and os.path.exists(data_args.validation_file_dir):
        eval_data_files = glob(f'{data_args.validation_file_dir}/**/*.txt', recursive=True)
        logger.info(f"eval files: {', '.join(eval_data_files)}")
        data_files["validation"] = eval_data_files

    print(data_files)
    raw_datasets = load_dataset(
        path=data_args.train_file_dir,
        cache_dir=training_args.cache_dir,
    )
    print(raw_datasets)
    raw_datasets = raw_datasets.shuffle(seed=42)

    # If no validation data is there, validation_split_percentage will be used to divide the dataset.
    if "validation" not in raw_datasets.keys():
        # raw_datasets["validation"] = load_dataset(
        #     path=data_args.train_file_dir,
        #     split=f"train[:{data_args.validation_split_percentage}%]",
        #     cache_dir=training_args.cache_dir,
        #     **dataset_args,
        # )
        # raw_datasets["train"] = load_dataset(
        #     path=data_args.train_file_dir,
        #     split=f"train[{data_args.validation_split_percentage}%:]",
        #     cache_dir=training_args.cache_dir,
        #     **dataset_args,
        # )
        # raw_datasets["validation"] = load_dataset(
        #     path=data_args.train_file_dir,
        #     split=f"train[:{data_args.validation_split_percentage}%]",
        #     cache_dir=training_args.cache_dir,
        #     **dataset_args,
        # )
        raw_datasets["train"] = load_dataset(
            path=data_args.train_file_dir,
            cache_dir=training_args.cache_dir,
            split=f"train[:]",
            **dataset_args,
        )
        raw_datasets["validation"] = load_dataset(
            path=data_args.train_file_dir,
            split=f"train[:2000]",
            cache_dir=training_args.cache_dir,
            **dataset_args,
        )
    logger.info(f"Raw datasets: {raw_datasets}")

    # Preprocessing the datasets.
    # if training_args.do_train:
    #     column_names = list(raw_datasets["train"].features)
    # else:
    #     column_names = list(raw_datasets["validation"].features)

    # column_names = list(raw_datasets["train"].features)

    with training_args.main_process_first(desc="Dataset tokenization and grouping"):
        if not data_args.streaming:
            tokenized_datasets = raw_datasets.map(
                tokenize_function,
                batched=True,
                num_proc=data_args.preprocessing_num_workers,
                remove_columns=["text", "idx"],
                load_from_cache_file=not data_args.overwrite_cache,
                desc="Running tokenizer on dataset",
            )
            lm_datasets = tokenized_datasets.map(
                group_texts,
                batched=True,
                num_proc=data_args.preprocessing_num_workers,
                load_from_cache_file=not data_args.overwrite_cache,
            )
        else:
            tokenized_datasets = raw_datasets.map(
                tokenize_function,
                batched=True,
                remove_columns=["text", "idx"],
            )
            lm_datasets = tokenized_datasets.map(
                group_texts,
                batched=True,
            )

    train_dataset = None
    max_train_samples = 0

    if "train" not in tokenized_datasets:
        raise ValueError("--do_train requires a train dataset")
    train_dataset = lm_datasets['train']
    print(train_dataset)
    max_train_samples = len(train_dataset)
    if data_args.max_train_samples is not None and data_args.max_train_samples > 0:
        max_train_samples = min(len(train_dataset), data_args.max_train_samples)
        train_dataset = train_dataset.select(range(max_train_samples))
    logger.debug(f"Num train_samples: {len(train_dataset)}")
    logger.debug("Tokenized training example:")
    logger.debug(len(train_dataset[0]['input_ids']))
    logger.debug(tokenizer.decode(train_dataset[0]['input_ids']))

    eval_dataset = None
    try:
        max_eval_samples = 0

        if "validation" not in tokenized_datasets:
            raise ValueError("--do_eval requires a validation dataset")
        eval_dataset = lm_datasets["validation"]
        max_eval_samples = len(eval_dataset)
        if data_args.max_eval_samples is not None and data_args.max_eval_samples > 0:
            max_eval_samples = min(len(eval_dataset), data_args.max_eval_samples)
            eval_dataset = eval_dataset.select(range(max_eval_samples))
        logger.debug(f"Num eval_samples: {len(eval_dataset)}")
        logger.debug("Tokenized eval example:")
        try:
            logger.debug(tokenizer.decode(eval_dataset[0]['input_ids']))
        except:
            logger.debug('no eval')
    except:
        print('no eval dataset')
    print(train_dataset)
    print(eval_dataset)

    return dict(train_dataset=train_dataset, eval_dataset=eval_dataset)





def train():
    global local_rank

    parser = transformers.HfArgumentParser(
        (ModelArguments, DataArguments, TrainingArguments)
    )
    model_args, data_args, training_args = parser.parse_args_into_dataclasses()
    local_rank = training_args.local_rank
    model = transformers.AutoModelForCausalLM.from_pretrained(
        model_args.model_name_or_path,
        cache_dir=training_args.cache_dir,
    )
    model.config.use_cache = False
    tokenizer = transformers.AutoTokenizer.from_pretrained(
        model_args.model_name_or_path,
        cache_dir=training_args.cache_dir,
        model_max_length=training_args.model_max_length,
        padding_side="right",
        use_fast=False,
    )
    tokenizer.pad_token = tokenizer.unk_token

    data_module = get_dataset(tokenizer, data_args, training_args)

    trainer = Trainer(
        model=model, tokenizer=tokenizer, args=training_args, **data_module
    )

    if list(pathlib.Path(training_args.output_dir).glob("checkpoint-*")):
        trainer.train(resume_from_checkpoint=True)
    else:
        trainer.train()
    trainer.save_state()
    safe_save_model_for_hf_trainer(trainer=trainer, output_dir=training_args.output_dir)




if __name__ == "__main__":
    train()
